@@ -1,4 +1,5 @@
 import gc
+import math
 import os
 import timeit
 
@@ -18,6 +19,13 @@ class ModelTorchAdapter(Model):
     useGPU = torch.cuda.is_available()
     useTensorBoard = False
     saveOnlyWeights = True
+    findLearningRate = False
+    oneCyclicLR = False
+    cycleLR = False
+    cycleLRFactor = 4
+    cycleLRStepsideFactor = 2
+    cycleLRMax = 2
+    findLearningRatePlot = "data/learningRate.png"
 
     def _metrics(self, metric):
         return "val_" + metric["name"]
@@ -140,6 +148,7 @@ class ModelTorchAdapter(Model):
 
         model = self.model
         globalBestMetric = self.validationMetrics[0]
+        self.batchScheduler = None
 
         self.log("LR: %s" % self.learningRate)
         if self.optimizer == "adams":
@@ -148,13 +157,34 @@ class ModelTorchAdapter(Model):
         if self.optimizer == "nadams":
             decay = 0 if self.learningRateDecay is None else self.learningRateDecay
             torch.optim.NAdam(model.parameters(), lr=self.learningRate, momentum_decay=decay)
-        elif self.optimizer == "adas":
-            from teleschlafmedizin.model.adapter.optimizer.Adas import Adas
-
-            self.optimizer = Adas(model.parameters(), lr=self.learningRate)
         elif type(self.optimizer) == str:
             raise Exception("optimizer %s is currently not supported for pytorch implementation" % self.optimizer)
 
+        if self.cycleLR:
+            lr = self.learningRate / self.cycleLRFactor
+            lrMax = self.learningRate
+            stepSize = self.cycleLRStepsideFactor * len(dataset.trainingData)
+
+            self.batchScheduler = torch.optim.lr_scheduler.CyclicLR(
+                self.optimizer, base_lr=self.learningRate / lr, max_lr=lrMax, step_size_up=stepSize
+            )
+
+        if self.oneCyclicLR:
+            lr = self.learningRate / self.cycleLRFactor
+            lrMax = self.learningRate
+            steps = self.maxEpochs * len(dataset.trainingData)
+
+            self.batchScheduler = torch.optim.lr_scheduler.OneCycleLR(self.optimizer, max_lr=lrMax, total_steps=steps)
+
+        if self.findLearningRate:
+            start_lr = self.learningRate
+            end_lr = self.cycleLRMax
+            lr_lambda = lambda x: math.exp(x * math.log(end_lr / start_lr) / (self.maxEpochs * len(dataset.trainingData)))
+            self.batchScheduler = torch.optim.lr_scheduler.LambdaLR(self.optimizer, lr_lambda)
+            lr_find_loss = []
+            lr_find_lr = []
+            smoothing = 0.05
+
         model = self.model
         lossCriterion = self.getLossFunction()
         i_epoch = self.startEpoch
@@ -211,6 +241,20 @@ class ModelTorchAdapter(Model):
                     lossCriterion(output, targs)
                     raise Exception("batch loss should not be a number")
 
+                if self.batchScheduler is not None:
+                    self.batchScheduler.step()
+
+                if self.findLearningRate:
+                    lr_step = self.optimizer.state_dict()["param_groups"][0]["lr"]
+                    lr_find_lr.append(lr_step)
+
+                    # smooth the loss
+                    if batchIndex == 0 and i_epoch == 0:
+                        lr_find_loss.append(loss)
+                    else:
+                        loss = smoothing * loss + (1 - smoothing) * lr_find_loss[-1]
+                        lr_find_loss.append(loss)
+
                 runningStats["loss"] += currentBatchLoss
 
                 del output
@@ -304,6 +348,20 @@ class ModelTorchAdapter(Model):
             if stopAfterNotImproving > 0 and notImprovedSince >= stopAfterNotImproving:
                 break
 
+        if self.findLearningRate:
+            lr_find_loss = [l.data.cpu().numpy() for l in lr_find_loss]
+            from matplotlib import pyplot as plt
+            plt.ylabel("loss")
+            plt.xlabel("loss")
+            plt.xscale("log")
+            plt.plot(lr_find_lr, lr_find_loss)
+            plt.savefig(self.findLearningRatePlotPath)
+            minLr = lr_find_lr[np.argmin(lr_find_loss)]
+            print("Lowest Loss LR: %f" % minLr)
+            print("Suggested LR range max bound: %f" % minLr)
+            self.lr_find_lr = lr_find_lr
+            self.lr_find_loss = lr_find_loss
+
         self.csvClose()
         self.fullEpochs = i_epoch
         return lastImprovdModel
@@ -349,7 +407,7 @@ class ModelTorchAdapter(Model):
         with torch.no_grad():
             if not torch.is_tensor(input):
                 input = torch.tensor(input)
- 
+
             batchSize, _, _ = input.shape
 
             if self.useGPU:
